from contextlib import contextmanager

from ..patterns import var_pattern


def replacer(match):
    gd = match.groupdict()
    if gd["dblquote"] is not None:
        return gd["dblquote"]
    elif gd["quote"] is not None:
        return gd["quote"]
    else:
        return f'{gd["lead"]}%({gd["var_name"]})s{gd["trail"]}'


class PsycoPG2Adapter:
    @staticmethod
    def process_sql(_query_name, _op_type, sql):
        return var_pattern.sub(replacer, sql)

    @staticmethod
    def select(conn, _query_name, sql, parameters, record_class=None):
        with conn.cursor() as cur:
            cur.execute(sql, parameters)
            results = cur.fetchall()
            if record_class is not None:
                column_names = [c.name for c in cur.description]
                results = [record_class(**dict(zip(column_names, row))) for row in results]
        return results

    @staticmethod
    def select_one(conn, _query_name, sql, parameters, record_class=None):
        with conn.cursor() as cur:
            cur.execute(sql, parameters)
            result = cur.fetchone()
            if result is not None and record_class is not None:
                column_names = [c.name for c in cur.description]
                result = record_class(**dict(zip(column_names, result)))
        return result

    @staticmethod
    @contextmanager
    def select_cursor(conn, _query_name, sql, parameters):
        with conn.cursor() as cur:
            cur.execute(sql, parameters)
            yield cur

    @staticmethod
    def insert_update_delete(conn, _query_name, sql, parameters):
        with conn.cursor() as cur:
            cur.execute(sql, parameters)

    @staticmethod
    def insert_update_delete_many(conn, _query_name, sql, parmeters):
        with conn.cursor() as cur:
            cur.executemany(sql, parmeters)

    @staticmethod
    def insert_returning(conn, _query_name, sql, parameters):
        with conn.cursor() as cur:
            cur.execute(sql, parameters)
            res = cur.fetchone()
            if res:
                return res[0] if len(res) == 1 else res
            else:
                return None

    @staticmethod
    def execute_script(conn, sql):
        with conn.cursor() as cur:
            cur.execute(sql)



from ..aioctxlib import aiocontextmanager


class AioSQLiteAdapter:
    is_aio_driver = True

    @staticmethod
    def process_sql(_query_name, _op_type, sql):
        """Pass through function because the ``aiosqlite`` driver can already handle the
        :var_name format used by aiosql and doesn't need any additional processing.

        Args:
            _query_name (str): The name of the sql query.
            _op_type (SQLOperationType): The type of SQL operation performed by the query.
            sql (str): The sql as written before processing.

        Returns:
            str: Original SQL text unchanged.
        """
        return sql

    @staticmethod
    async def select(conn, _query_name, sql, parameters, record_class=None):
        async with conn.execute(sql, parameters) as cur:
            results = await cur.fetchall()
            if record_class is not None:
                column_names = [c[0] for c in cur.description]
                results = [record_class(**dict(zip(column_names, row))) for row in results]
        return results

    @staticmethod
    async def select_one(conn, _query_name, sql, parameters, record_class=None):
        async with conn.execute(sql, parameters) as cur:
            result = await cur.fetchone()
            if result is not None and record_class is not None:
                column_names = [c[0] for c in cur.description]
                result = record_class(**dict(zip(column_names, result)))
        return result

    @staticmethod
    @aiocontextmanager
    async def select_cursor(conn, _query_name, sql, parameters):
        async with conn.execute(sql, parameters) as cur:
            yield cur

    @staticmethod
    async def insert_returning(conn, _query_name, sql, parameters):
        async with conn.execute(sql, parameters) as cur:
            return cur.lastrowid

    @staticmethod
    async def insert_update_delete(conn, _query_name, sql, parameters):
        cur = await conn.execute(sql, parameters)
        await cur.close()

    @staticmethod
    async def insert_update_delete_many(conn, _query_name, sql, parameters):
        cur = await conn.executemany(sql, parameters)
        await cur.close()

    @staticmethod
    async def execute_script(conn, sql):
        await conn.executescript(sql)



from collections import defaultdict

from ..aioctxlib import aiocontextmanager
from ..patterns import var_pattern


class MaybeAcquire:
    def __init__(self, client):
        self.client = client

    async def __aenter__(self):
        if "acquire" in dir(self.client):
            self._managed_conn = await self.client.acquire()
            return self._managed_conn
        else:
            self._managed_conn = None
            return self.client

    async def __aexit__(self, exc_type, exc, tb):
        if self._managed_conn is not None:
            await self.client.release(self._managed_conn)


class AsyncPGAdapter:
    is_aio_driver = True

    def __init__(self):
        self.var_replacements = defaultdict(dict)

    def process_sql(self, query_name, _op_type, sql):
        count = 0
        adj = 0

        for match in var_pattern.finditer(sql):
            gd = match.groupdict()
            # Do nothing if the match is found within quotes.
            if gd["dblquote"] is not None or gd["quote"] is not None:
                continue

            var_name = gd["var_name"]
            if var_name in self.var_replacements[query_name]:
                replacement = f"${self.var_replacements[query_name][var_name]}"
            else:
                count += 1
                replacement = f"${count}"
                self.var_replacements[query_name][var_name] = count

            start = match.start() + len(gd["lead"]) + adj
            end = match.end() - len(gd["trail"]) + adj

            sql = sql[:start] + replacement + sql[end:]

            replacement_len = len(replacement)
            # the lead ":" char is the reason for the +1
            var_len = len(var_name) + 1
            if replacement_len < var_len:
                adj = adj + replacement_len - var_len
            else:
                adj = adj + var_len - replacement_len

        return sql

    def maybe_order_params(self, query_name, parameters):
        if isinstance(parameters, dict):
            xs = [(self.var_replacements[query_name][k], v) for k, v in parameters.items()]
            xs = sorted(xs, key=lambda x: x[0])
            return [x[1] for x in xs]
        elif isinstance(parameters, tuple):
            return parameters
        else:
            raise ValueError(f"Parameters expected to be dict or tuple, received {parameters}")

    async def select(self, conn, query_name, sql, parameters, record_class=None):
        parameters = self.maybe_order_params(query_name, parameters)
        async with MaybeAcquire(conn) as connection:
            results = await connection.fetch(sql, *parameters)
            if record_class is not None:
                results = [record_class(**dict(rec)) for rec in results]
        return results

    async def select_one(self, conn, query_name, sql, parameters, record_class=None):
        parameters = self.maybe_order_params(query_name, parameters)
        async with MaybeAcquire(conn) as connection:
            result = await connection.fetchrow(sql, *parameters)
            if result is not None and record_class is not None:
                result = record_class(**dict(result))
        return result

    @aiocontextmanager
    async def select_cursor(self, conn, query_name, sql, parameters):
        parameters = self.maybe_order_params(query_name, parameters)
        async with MaybeAcquire(conn) as connection:
            stmt = await connection.prepare(sql)
            async with connection.transaction():
                yield stmt.cursor(*parameters)

    async def insert_returning(self, conn, query_name, sql, parameters):
        parameters = self.maybe_order_params(query_name, parameters)
        async with MaybeAcquire(conn) as connection:
            res = await connection.fetchrow(sql, *parameters)
            if res:
                return res[0] if len(res) == 1 else res
            else:
                return None

    async def insert_update_delete(self, conn, query_name, sql, parameters):
        parameters = self.maybe_order_params(query_name, parameters)
        async with MaybeAcquire(conn) as connection:
            await connection.execute(sql, *parameters)

    async def insert_update_delete_many(self, conn, query_name, sql, parameters):
        parameters = [self.maybe_order_params(query_name, params) for params in parameters]
        async with MaybeAcquire(conn) as connection:
            await connection.executemany(sql, parameters)

    @staticmethod
    async def execute_script(conn, sql):
        async with MaybeAcquire(conn) as connection:
            await connection.execute(sql)



from types import MethodType
from typing import Callable, List, Tuple

from .types import QueryDatum, QueryDataTree, SQLOperationType


def _create_methods(query_datum: QueryDatum, is_aio=True) -> List[Tuple[str, Callable]]:
    query_name, doc_comments, operation_type, sql, record_class = query_datum

    if is_aio:

        async def fn(self, conn, *args, **kwargs):
            parameters = kwargs if len(kwargs) > 0 else args
            if operation_type == SQLOperationType.INSERT_RETURNING:
                return await self.driver_adapter.insert_returning(conn, query_name, sql, parameters)
            elif operation_type == SQLOperationType.INSERT_UPDATE_DELETE:
                return await self.driver_adapter.insert_update_delete(
                    conn, query_name, sql, parameters
                )
            elif operation_type == SQLOperationType.INSERT_UPDATE_DELETE_MANY:
                return await self.driver_adapter.insert_update_delete_many(
                    conn, query_name, sql, *parameters
                )
            elif operation_type == SQLOperationType.SCRIPT:
                return await self.driver_adapter.execute_script(conn, sql)
            elif operation_type == SQLOperationType.SELECT:
                return await self.driver_adapter.select(
                    conn, query_name, sql, parameters, record_class
                )
            elif operation_type == SQLOperationType.SELECT_ONE:
                return await self.driver_adapter.select_one(
                    conn, query_name, sql, parameters, record_class
                )
            else:
                raise ValueError(f"Unknown op_type: {operation_type}")

    else:

        def fn(self, conn, *args, **kwargs):
            parameters = kwargs if len(kwargs) > 0 else args
            if operation_type == SQLOperationType.INSERT_RETURNING:
                return self.driver_adapter.insert_returning(conn, query_name, sql, parameters)
            elif operation_type == SQLOperationType.INSERT_UPDATE_DELETE:
                return self.driver_adapter.insert_update_delete(conn, query_name, sql, parameters)
            elif operation_type == SQLOperationType.INSERT_UPDATE_DELETE_MANY:
                return self.driver_adapter.insert_update_delete_many(
                    conn, query_name, sql, *parameters
                )
            elif operation_type == SQLOperationType.SCRIPT:
                return self.driver_adapter.execute_script(conn, sql)
            elif operation_type == SQLOperationType.SELECT:
                return self.driver_adapter.select(conn, query_name, sql, parameters, record_class)
            elif operation_type == SQLOperationType.SELECT_ONE:
                return self.driver_adapter.select_one(
                    conn, query_name, sql, parameters, record_class
                )
            else:
                raise ValueError(f"Unknown op_type: {operation_type}")

    fn.__name__ = query_name
    fn.__doc__ = doc_comments
    fn.sql = sql

    ctx_mgr_method_name = f"{query_name}_cursor"

    def ctx_mgr(self, conn, *args, **kwargs):
        parameters = kwargs if len(kwargs) > 0 else args
        return self.driver_adapter.select_cursor(conn, query_name, sql, parameters)

    ctx_mgr.__name__ = ctx_mgr_method_name
    ctx_mgr.__doc__ = doc_comments
    ctx_mgr.sql = sql

    if operation_type == SQLOperationType.SELECT:
        return [(query_name, fn), (ctx_mgr_method_name, ctx_mgr)]
    else:
        return [(query_name, fn)]


class Queries:
    """Container object with dynamic methods built from SQL queries.

    The ``-- name`` definition comments in the SQL content determine what the dynamic
    methods of this class will be named.

    @DynamicAttrs
    """

    def __init__(self, driver_adapter):
        """Queries constructor.

        Args:
            driver_adapter (object):
        """
        self.driver_adapter = driver_adapter
        self.is_aio = getattr(driver_adapter, "is_aio_driver", False)
        self._available_queries = set()

    @property
    def available_queries(self):
        """Returns listing of all the available query methods loaded in this class.

        Returns:
            list(str): List of dot-separated method accessor names.
        """
        return sorted(self._available_queries)

    def __repr__(self):
        return "Queries(" + self.available_queries.__repr__() + ")"

    def add_query(self, query_name, fn):
        """Adds a new dynamic method to this class.

        Args:
            query_name (str): The method name as found in the SQL content.
            fn (function): The loaded query function.

        Returns:

        """
        setattr(self, query_name, fn)
        self._available_queries.add(query_name)

    def add_queries(self, queries):
        for query_name, fn in queries:
            self.add_query(query_name, MethodType(fn, self))

    def add_child_queries(self, child_name, child_queries):
        """Adds a Queries object as a property.

        Args:
            child_name (str): The property name to group the child queries under.
            child_queries (Queries): Queries instance to add as sub-queries.

        Returns:
            None

        """
        setattr(self, child_name, child_queries)
        for child_query_name in child_queries.available_queries:
            self._available_queries.add(f"{child_name}.{child_query_name}")

    def load_from_list(self, query_data: List[QueryDatum]):
        for query_datum in query_data:
            self.add_queries(_create_methods(query_datum, self.is_aio))
        return self

    def load_from_tree(self, query_data_tree: QueryDataTree):
        for key, value in query_data_tree.items():
            if isinstance(value, dict):
                self.add_child_queries(key, Queries(self.driver_adapter).load_from_tree(value))
            else:
                self.add_queries(_create_methods(value, self.is_aio))
        return self



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['aiosql', 'aiosql.adapters']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'aiosql',
    'version': '3.1.0',
    'description': 'Simple SQL in Python.',
    'long_description': '# aiosql\n\nSimple SQL in Python.\n\nSQL is code, you should be able to write it, version control it, comment on it, and use it in database tools\nlike `psql` as you would any other SQL. But, you also want to be able to use it from your python\napplications, and that\'s where `aiosql` can help. With `aiosql` you can organize your SQL statements in `.sql`\nfiles and load them into a python object as methods to call.\n\nThis project supports sync and asyncio based drivers for SQLite (`sqlite3`, `aiosqlite`) and PostgreSQL\n(`psycopg2`, `asyncpg`) out of the box, and can be extended to support other database drivers by you! The ``asyncio``\nsupport restricts this package to python versions >3.6. If you are using older versions of python please see the\nrelated [anosql](https://github.com/honza/anosql) package which this project is based on.\n\n## Install\n\n```\npip install aiosql\n```\n\nOr if you you use [poetry](https://poetry.eustace.io/):\n\n```\npoetry add aiosql\n```\n\n## Getting Started\n\n#### Basic Usage\n\nGiven you have a SQL file like the one below called `users.sql`\n\n```sql\n-- name: get-all-users\n-- Get all user records\nselect * from users;\n\n\n-- name: get-user-by-username\n-- Get user with the given username field.\nselect userid,\n       username,\n       firstname,\n       lastname\n  from users\n where username = :username;\n```\n\nYou can use `aiosql` to load the queries in this file for use in your Python application:\n\n```python\nimport aiosql\nimport sqlite3\n\nconn = sqlite3.connect("myapp.db")\nqueries = aiosql.from_path("users.sql", "sqlite3")\n\nusers = queries.get_all_users(conn)\n# >>> [(1, "nackjicholson", "William", "Vaughn"), (2, "johndoe", "John", "Doe"), ...]\n\nusers = queries.get_user_by_username(conn, username="nackjicholson")\n# >>> [(1, "nackjicholson", "William", "Vaughn")\n```\n\nThis is pretty nice, we\'re able to define our methods in SQL and use them as methods from python!\n\n#### Query Operators to define different types of SQL actions\n\n`aiosql` can help you do even more by allowing you to declare in the SQL how you would like a query to be executed\nand returned in python. For instance, the `get-user-by-username` query above should really only return a single result\ninstead of a list containing one user. With the raw `sqlite3` driver in python we would probably have used \n`cur.fetchone()` instead of `cur.fetchall()` to retrieve a single row. We can inform `aiosql` to select a single row\nby using the `^` (select one) operator on the end of our query name.\n\n```sql\n-- name: get-user-by-username^\n-- Get user with the given username field.\nselect userid,\n       username,\n       firstname,\n       lastname\n  from users\n where username = :username;\n```\n\n```python\nnack = queries.get_user_by_username(conn, username="nackjicholson")\n# >>> (1, "nackjicholson", "William", "Vaughn")\n```\n\n#### Using your own python types for SQL data.\n\nBy declaring a `record_class` directive in our SQL file we can inform `aiosql` to automatically marshal our data to a\ncustom class we\'ve defined in python. In python3.7 a good choice for this is the new `dataclass` package.\n\n```sql\n-- name: get-user-by-username^\n-- record_class: User\n-- Get user with the given username field.\nselect userid,\n       username,\n       firstname,\n       lastname\n  from users\n where username = :username;\n```\n\nAll we have to do is provide our custom type to `aiosql` when we load our queries via the `record_classes` argument.\n\n```python\nimport aiosql\nimport sqlite3\nfrom dataclasses import dataclass\n\n\n@dataclass\nclass User:\n    userid: int\n    username: str\n    firstname: str\n    lastname: str\n\n\nconn = sqlite3.connect("myapp.db")\nqueries = aiosql.from_path("users.sql", "sqlite3", record_classes={"User": User})\n\nnack = queries.get_user_by_username(conn, username="nackjicholson")\n# >>> User(userid=1, username="nackjicholson", firstname="William", lastname="Vaughn")\n```\n\nHopefully this is enough to intrigue you and entice you to give aiosql a try. Check the documentation site for more\ninformation, and more features. Happy SQLing!\n\n## Documentation\n\nProject and API docs https://nackjicholson.github.io/aiosql\n',
    'author': 'William Vaughn',
    'author_email': 'vaughnwilld@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/nackjicholson/aiosql',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



from pathlib import Path
from typing import Dict, List, Optional

from .exceptions import SQLParseException, SQLLoadException
from .patterns import (
    doc_comment_pattern,
    empty_pattern,
    query_record_class_definition_pattern,
    query_name_definition_pattern,
    valid_query_name_pattern,
)
from .types import QueryDatum, QueryDataTree, SQLOperationType


class QueryLoader:
    def __init__(self, driver_adapter, record_classes: Optional[Dict]):
        self.driver_adapter = driver_adapter
        self.record_classes = record_classes if record_classes is not None else {}

    def _make_query_datum(self, query_str: str):
        lines = query_str.strip().splitlines()
        query_name = lines[0].replace("-", "_")

        if query_name.endswith("<!"):
            operation_type = SQLOperationType.INSERT_RETURNING
            query_name = query_name[:-2]
        elif query_name.endswith("*!"):
            operation_type = SQLOperationType.INSERT_UPDATE_DELETE_MANY
            query_name = query_name[:-2]
        elif query_name.endswith("!"):
            operation_type = SQLOperationType.INSERT_UPDATE_DELETE
            query_name = query_name[:-1]
        elif query_name.endswith("#"):
            operation_type = SQLOperationType.SCRIPT
            query_name = query_name[:-1]
        elif query_name.endswith("^"):
            operation_type = SQLOperationType.SELECT_ONE
            query_name = query_name[:-1]
        else:
            operation_type = SQLOperationType.SELECT

        if not valid_query_name_pattern.match(query_name):
            raise SQLParseException(
                f'name must convert to valid python variable, got "{query_name}".'
            )

        record_class_match = query_record_class_definition_pattern.match(lines[1])
        if record_class_match:
            line_offset = 2
            record_class_name = record_class_match.group(1)
        else:
            line_offset = 1
            record_class_name = None

        doc_comments = ""
        sql = ""
        for line in lines[line_offset:]:
            doc_match = doc_comment_pattern.match(line)
            if doc_match:
                doc_comments += doc_match.group(1) + "\n"
            else:
                sql += line + "\n"

        doc_comments = doc_comments.strip()
        sql = self.driver_adapter.process_sql(query_name, operation_type, sql.strip())
        record_class = self.record_classes.get(record_class_name)

        return QueryDatum(query_name, doc_comments, operation_type, sql, record_class)

    def load_query_data_from_sql(self, sql: str) -> List[QueryDatum]:
        query_data = []
        for query_sql_str in query_name_definition_pattern.split(sql):
            if not empty_pattern.match(query_sql_str):
                query_data.append(self._make_query_datum(query_sql_str))
        return query_data

    def load_query_data_from_file(self, file_path: Path) -> List[QueryDatum]:
        with file_path.open() as fp:
            return self.load_query_data_from_sql(fp.read())

    def load_query_data_from_dir_path(self, dir_path) -> QueryDataTree:
        if not dir_path.is_dir():
            raise ValueError(f"The path {dir_path} must be a directory")

        def _recurse_load_query_data_tree(path):
            # queries = Queries()
            query_data_tree = {}
            for p in path.iterdir():
                if p.is_file() and p.suffix != ".sql":
                    continue
                elif p.is_file() and p.suffix == ".sql":
                    for query_datum in self.load_query_data_from_file(p):
                        query_data_tree[query_datum.query_name] = query_datum
                elif p.is_dir():
                    child_name = p.relative_to(dir_path).name
                    child_query_data_tree = _recurse_load_query_data_tree(p)
                    query_data_tree[child_name] = child_query_data_tree
                else:
                    # This should be practically unreachable.
                    raise SQLLoadException(f"The path must be a directory or file, got {p}")
            return query_data_tree

        return _recurse_load_query_data_tree(dir_path)



from pathlib import Path
from typing import Dict, Union, Any, Optional, Type

from .adapters.aiosqlite import AioSQLiteAdapter
from .adapters.asyncpg import AsyncPGAdapter
from .adapters.psycopg2 import PsycoPG2Adapter
from .adapters.sqlite3 import SQLite3DriverAdapter
from .exceptions import SQLLoadException

from .queries import Queries
from .query_loader import QueryLoader


_ADAPTERS = {
    "aiosqlite": AioSQLiteAdapter,
    "asyncpg": AsyncPGAdapter,
    "psycopg2": PsycoPG2Adapter,
    "sqlite3": SQLite3DriverAdapter,
}


def _make_driver_adapter(driver_adapter: Union[str, Any]):
    """Get the driver adapter instance registered by the ``driver_name``.

    Args:
        driver_adapter (str|Any): The database driver name.

    Returns:
        object: A driver adapter class.
    """
    if isinstance(driver_adapter, str):
        try:
            driver_adapter = _ADAPTERS[driver_adapter]
        except KeyError:
            raise ValueError(f"Encountered unregistered driver_adapter: {driver_adapter}")

    return driver_adapter()


def from_str(
    sql,
    driver_adapter,
    record_classes=None,
    *,
    loader_cls: Type[QueryLoader] = QueryLoader,
    queries_cls: Type[Queries] = Queries,
):
    """Load queries from a SQL string.

    Args:
        sql (str) A string containing SQL statements and aiosql name:
        driver_adapter (str|Any): Either a string to designate one of the aiosql built-in
                                  database driver adapters. One of "sqlite3", "psycopg2",
                                  "aiosqlite", or "asyncpg". If you have defined your own adapter
                                  class, you can pass it's constructor.
        record_classes (dict|None): Mapping of strings used in "record_class" declarations to the
                                    python classes which aiosql should use when marshaling SQL
                                    results.
        loader_cls (QueryLoader): Custom constructor for QueryLoader extensions.
        queries_cls (Queries): Custom constructor for Queries extensions.

    Returns:
        Queries

    Example:
        Loading queries from a SQL string::

            import sqlite3
            import aiosql

            sql_text = \"""
            -- name: get-all-greetings
            -- Get all the greetings in the database
            select * from greetings;

            -- name: get-user-by-username^
            -- Get all the users from the database,
            -- and return it as a dict
            select * from users where username =:username;
            \"""

            queries = aiosql.from_str(sql_text, "sqlite3")
            queries.get_all_greetings(conn)
            queries.get_user_by_username(conn, username="willvaughn")

    """
    driver_adapter = _make_driver_adapter(driver_adapter)
    query_loader = loader_cls(driver_adapter, record_classes)
    query_data = query_loader.load_query_data_from_sql(sql)
    return queries_cls(driver_adapter).load_from_list(query_data)


def from_path(
    sql_path: Union[str, Path],
    driver_adapter: Union[str, Any],
    record_classes: Optional[Dict] = None,
    *,
    loader_cls: Type[QueryLoader] = QueryLoader,
    queries_cls: Type[Queries] = Queries,
):
    """Load queries from a ``.sql`` file, or directory of ``.sql`` files.

    Args:
        sql_path (str|Path): Path to a ``.sql`` file or directory containing ``.sql`` files.
        driver_adapter (str|Any): Either a string to designate one of the aiosql built-in
                                  database driver adapters. One of "sqlite3", "psycopg2",
                                  "aiosqlite", or "asyncpg". If you have defined your own adapter
                                  class, you can pass it's constructor.
        record_classes (dict|None): Mapping of strings used in "record_class" declarations to the
                                    python classes which aiosql should use when marshaling SQL
                                    results.
        loader_cls (QueryLoader): Custom constructor for QueryLoader extensions.
        queries_cls (Queries): Custom constructor for Queries extensions.

    Returns:
        Queries: Queries object.

    Example::

        >>> aiosql.from_path("./sql", "pscycopg2")
        >>> aiosql.from_path("./sql", MyDBAdapter)
        >>> aiosql.from_path("./sql/users.sql", "aiosqlite", record_classes={"User": User})

    """
    path = Path(sql_path)

    if not path.exists():
        raise SQLLoadException(f"File does not exist: {path}")

    driver_adapter = _make_driver_adapter(driver_adapter)
    query_loader = loader_cls(driver_adapter, record_classes)

    if path.is_file():
        query_data = query_loader.load_query_data_from_file(path)
        return queries_cls(driver_adapter).load_from_list(query_data)
    elif path.is_dir():
        query_data_tree = query_loader.load_query_data_from_dir_path(path)
        return queries_cls(driver_adapter).load_from_tree(query_data_tree)
    else:
        raise SQLLoadException(f"The sql_path must be a directory or file, got {sql_path}")



class SQLLoadException(Exception):
    """Raised when there is a problem loading SQL content from a file or directory."""

    pass


class SQLParseException(Exception):
    """Raised when there was a problem parsing the aiosql comment annotations in SQL"""

    pass



from enum import Enum
from typing import Dict, Union, NamedTuple, Any


class SQLOperationType(Enum):
    """Enumeration of aiosql operation types.
    """

    INSERT_RETURNING = 0
    INSERT_UPDATE_DELETE = 1
    INSERT_UPDATE_DELETE_MANY = 2
    SCRIPT = 3
    SELECT = 4
    SELECT_ONE = 5


class QueryDatum(NamedTuple):
    query_name: str
    doc_comments: str
    operation_type: SQLOperationType
    sql: str
    record_class: Any = None


QueryDataTree = Dict[str, Union[QueryDatum, "QueryDataTree"]]
